function angle(a, b, c) = acos((b*b + c*c - a*a)/(2*b*c));

function angle_between(v0, v1) = acos((v0 * v1) / (norm(v0) * norm(v1)));

function length(p0, p1) = sqrt((p0[0] - p1[0])*(p0[0] - p1[0]) + (p0[1] - p1[1])*(p0[1] - p1[1]));

function triangle_center(p0, p1, p2) =
    let(L0 = length(p0, p1))
    let(L1 = length(p1, p2))
    let(L2 = length(p2, p0))
    let(alpha=angle(L2, L0, L1))
    let(beta=angle(L1, L0, L2))
    [L0*L1*(2*L0 - L2*cos(beta))*sin(alpha)/(4*L0*L1*sin(alpha) + L0*L2*sin(beta) - 2*L1*L2*sin(alpha + beta)),
             L0*L1*L2*sin(alpha)*sin(beta)/(4*L0*L1*sin(alpha) + L0*L2*sin(beta) - 2*L1*L2*sin(alpha + beta))];

function triangle_center2(p0,p1,p2) =
    (p0 + p1 + p2) / 3;

function normal(v0, v1) =
    cross(v0, v1) / norm(cross(v0, v1));

function skewMatrix(v) =
//Create a skew antisymmetric matrix from a vector
    [[0,    -v[2],  v[1]],
     [v[2] , 0   , -v[0]],
     [-v[1], v[0], 0    ]];

function rotAxis(a, b) =
//Find the rotation axis along which to point vector a along b
    cross(a, b) / norm(cross(a, b));

function rotAngle(a, b) =
//Find the rotation angle between vector a and b
    acos((a * b) / (norm(a) * norm(b)));

function rotMatrix(a, b) =
//Construct the rotation matrix that rotates vector a to lie along b
    let(I = [[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    let(x = rotAxis(a, b))
    let(A = skewMatrix(x))
    let(th = rotAngle(a, b))
    I + sin(th) * A + (1 - cos(th)) * A * A;

module triangle(p0, p1, p2, t){
    dt = t * normal((p1-p0),    (p2-p0));
    polyhedron([p0, p1, p2, p0+dt, p1+dt, p2+dt],
               [[0,1,2],
                [3, 4, 5],
                [0, 1, 4, 3],
                [1, 2, 5, 4],
                [2, 0, 3, 5]]);
}

module triangle2(p0, p1, p2, t, d=.01, fn=20){
    hull(){
        translate(p0){
            cylinder(d=d, h=t, center=true, $fn=fn);
        }
        translate(p1){
            cylinder(d=d, h=t, center=true, $fn=fn);
        }
        translate(p2){
            cylinder(d=d, h=t, center=true, $fn=fn);
        }
    }
}

module bar(p0, p1, t, fn=20){
    hull(){
        translate(p0) cylinder(d=t, h=t, center=true, $fn=fn);
        translate(p1) cylinder(d=t, h=t, center=true, $fn=fn);
    }
}

module tile(p0, p1, p2, theta, t, s=.8, fn=20){
    //Create a 2D tile, placed in the x-y plane
    //p0, p1, p2 are the 2D coordinates [x,y] of the corners
    //theta is the rotation angle of the cell
    //t is the thickness of the plates
    //s is the scale wich decreases the size of the actual plates
    //fn (optional) is the number of faces used to approximate a sphere
    disk_t = .2;  // Thickness (fraction) of disks on both sides of joints
    joint_scale = 1;  // Scale of joint radius (in terms of thickness)
    a = norm(p0 - p1) / (2 * cos(theta));
    b = norm(p2 - p0) / (2 * cos(theta));
    c = norm(p2 - p1) / (2 * cos(theta));
    cent = triangle_center2(p0,p1,p2);
    alpha = angle_between((p1-p2), (p0-p2));
    beta = angle_between((p2-p1), (p0-p1));
    gamma = angle_between((p1-p0), (p2-p0));
    p3 = p0 + a * [cos(theta), sin(theta)];
    p4 = p1 + c * [-cos(beta + theta), sin(beta + theta)];
    p5 = p0 + b * [cos(theta + gamma), sin(theta + gamma)];

    difference(){
        union(){  //Rigid parts
            //Outer points
            for (p=[p0,p1,p2]){
                hull(){
                    translate(p) sphere(d=t, $fn=fn);
                    translate(p + (1-s) * (cent-p)) cylinder(d=s*t, h=t, $fn=fn, center=true);
                }
            }
            //Scale with respect to centroid
            translate(triangle_center2(p0, p1, p2))
            scale([s, s, 1.])
            translate(-triangle_center2(p0, p1, p2)){
                triangle2(p0, p3, p5, t);
                triangle2(p1, p3, p4, t);
                bar(p5, p2, t);
                bar(p4, p2, t);
            }
            translate(triangle_center2(p0, p1, p2))
            scale([s, s, 1.])
            translate(-triangle_center2(p0, p1, p2)){
                union(){  // Cutouts for soft parts
                    for (p=[p0,p1,p2,p3,p4,p5]){
                        translate([0,0,-t/2]) translate(p)
                        cylinder(d=1.*t, h=t, $fn=fn);
                    }
                }
            }
        }
    }
}


module place_tile(p0, p1, p2, theta, t, s=.8, hinge, fn=20){
    //Place a triangular tile in space, such that the corners align with points
    //p0, p1, p2. The other parameters are fed throug to the tile module.
    //echo(hinge);
    //Define z-direction
    z = [0,0,1];
    //Convert vectors to 2D and translate to origin
    P0 = [0,0];
    P1 = norm(p1-p0) * [1, 0];
    P2 = norm(p2-p0) * [cos(angle_between((p1-p0), (p2-p0))), sin(angle_between((p1-p0), (p2-p0)))];
    //Normal
    n = normal(p1-p0, p2-p0);
    //Translate to the first point
	translate(p0)
    //Rotate to align the normal of the triangle with the calculated normal
    rotate(a=rotAngle(z,n), v=rotAxis(z,n)){
        //Rotate in-plane to align the edge p0-p1
        if (z == n){
            v1 = [P1[0], P1[1], 0];
            v2 = (p1 - p0);
            if ((v1==-v2)){
                rotate(a=180, v=z){
                    //Construct a triangular tile in-plane
                    tile(P0, P1, P2, theta, t, s=s, hinge);
                }
            }else {
                rotate(a=rotAngle(v1,v2), v=rotAxis(v1, v2)){
                    //Construct a triangular tile in-plane
                    tile(P0, P1, P2, theta, t, s=s, hinge=hinge);
                }
            }
        } else if (z == -n){
            v1 = [P1[0], P1[1], 0];
            v2 = (p1 - p0);
            rotate(a=180, v=v2)
            if ((v1==-v2)){
                rotate(a=180, v=z){
                    //Construct a triangular tile in-plane
                    tile(P0, P1, P2, theta, t, s=s, hinge=hinge);
                }
            }else {
                rotate(a=rotAngle(v1,v2), v=rotAxis(v1, v2)){
                    //Construct a triangular tile in-plane
                    tile(P0, P1, P2, theta, t, s=s, hinge=hinge);
                }
            }
        }else{
            v1 = [P1[0], P1[1], 0];
            v2 = (p1 - p0) * rotMatrix(z,n);
            if ((v1==-v2)){
                rotate(a=180, v=z){
                    //Construct a triangular tile in-plane
                    tile(P0, P1, P2, theta, t, s=s, hinge=hinge);
                }
            }else {
                rotate(a=rotAngle(v1,v2), v=rotAxis(v1, v2)){
                    //Construct a triangular tile in-plane
                    tile(P0, P1, P2, theta, t, s=s, hinge=hinge);
                }
            }
        }
    }
}
